#include "UdpDevice.h"

#include <QUdpSocket>
#include <QIODevice>
#include <QMetaMethod>
#include <QTime>

struct Client
{
		QHostAddress adress;
		quint16 port;
};

UdpDevice::UdpDevice():
	client(0)
{
}

UdpDevice::UdpDevice(QString port, int param):
	client(0)
{
	initiate(port, param);
}

UdpDevice::~UdpDevice()
{
	close();
}

int UdpDevice::write(const QByteArray& data)
{
	if(isOpen())
	{
		if(isServerMode())
			return socket->writeDatagram(data, client->adress, client->port);
		else
			return socket->writeDatagram(data, socket->peerAddress(), socket->peerPort());
	}
	else
		return -1;
}


bool UdpDevice::initiate(QString port, int param)
{
	if(socket.isNull())
	{
		socket.reset(new QUdpSocket(this));
		connect(socket.data(), &QUdpSocket::connected, this, &UdpDevice::portOpened);
		connect(socket.data(), &QUdpSocket::disconnected, this, &UdpDevice::portClosed);
		connect(socket.data(), &QUdpSocket::readyRead, this, &UdpDevice::readyRead);
		connect(socket.data(), &QUdpSocket::readyRead, [this]() { qDebug() << "ReadyRead emited for " << socket->localPort() << " : " << socket->hasPendingDatagrams(); });
		connect(socket.data(), &QUdpSocket::stateChanged, [](QAbstractSocket::SocketState state) { qDebug() << "Udp state changed : " << state; });
		connect(socket.data(), static_cast<void(QUdpSocket::*)(QUdpSocket::SocketError)>(&QUdpSocket::error), [this]() { emit errorOccurred(socket->errorString()); });

		if(isServerMode())
			delete client;
		client = 0;
	}

	if(socket.isNull())
		return false;
	else if(param != 0)
	{
		if(param < 0)
		{
			//	Mode serveur
			client = new Client;
			if(socket->bind(QHostAddress::LocalHost, -param, QAbstractSocket::DontShareAddress))
				emit portOpened();
			else
				emit errorOccurred(tr("Failed to open socket port (server mode) %1 :\n\t%2").arg(-param).arg(socket->errorString()));

			return socket->state() == QUdpSocket::BoundState;
		}
		else
		{
			//	Mode client
			socket->connectToHost(port.isEmpty() ? QHostAddress(QHostAddress::LocalHost) : QHostAddress(port), param);
			if(socket->waitForConnected())
				emit portOpened();
			else
				emit errorOccurred(tr("Failed to open socket port %1 :\n\t%2").arg(param).arg(socket->errorString()));

			return socket->state() == QUdpSocket::ConnectedState;
		}
	}
	else
	{
		emit errorOccurred(tr("Cannot open a Udp socket on the port 0."));
		return false;
	}
}

bool UdpDevice::isOpen() const
{
	return !socket.isNull() && (isServerMode() ? socket->state() == QUdpSocket::BoundState : socket->state() == QUdpSocket::ConnectedState);
}

bool UdpDevice::isInitiated() const
{
	return !socket.isNull() && socket->state() != QUdpSocket::UnconnectedState;
}

void UdpDevice::close()
{
	if(isOpen())
		socket->close();
}

QByteArray UdpDevice::readAll()
{
	QByteArray data;
	if(isOpen())
	{
		qDebug() << "Read from " << socket->localPort();
		if(socket->hasPendingDatagrams())
		{
			QByteArray res;
			res.resize(socket->pendingDatagramSize());
			int ret(0);
			if(isServerMode())
				ret = socket->readDatagram(res.data(), res.size(), &client->adress, &client->port);
			else
				ret = socket->readDatagram(res.data(), res.size());

			qDebug() << "Datagram : " << res;
			data = res;

			if(-1 == ret)
			{
				emit errorOccurred(tr("Error when reading Udp Socket : ") + socket->errorString());
			}
		}
		else
			emit errorOccurred("No datagram to read");
	}
	return data;
}


QString UdpDevice::errorString() const
{
	if(socket.isNull())
		return QString();
	else
		return socket->errorString();
}

bool UdpDevice::waitForReadyRead(int msec)
{
	if(isOpen())
		return socket->waitForReadyRead(msec);
	else
		return false;
}

QString UdpDevice::portName() const
{
	if(socket.isNull())
		return QString();
	else
		return QString::number(isServerMode() ? socket->localPort() : socket->peerPort());
}

QString UdpDevice::stateString() const
{
	if(socket.isNull())
		return "No Udp socket initated.";
	else
		return QVariant(socket->state()).toString();
}

bool UdpDevice::isServerMode() const
{
	return client != 0;
}
